'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = events;

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _helpers = require('./helpers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Graphology Events Specs
 * ========================
 *
 * Testing the graph's events.
 */
var VALID_TYPES = new Set(['set', 'merge', 'replace']);

function events(Graph) {
  return {
    'nodeAdded': {
      'it should fire when a node is added.': function itShouldFireWhenANodeIsAdded() {
        var graph = new Graph();

        var handler = (0, _helpers.spy)(function (data) {
          _assert2.default.strictEqual(data.key, 'John');
          _assert2.default.deepEqual(data.attributes, { age: 34 });
        });

        graph.on('nodeAdded', handler);

        graph.addNode('John', { age: 34 });

        (0, _assert2.default)(handler.called);
      }
    },

    'edgeAdded': {
      'it should fire when an edge is added.': function itShouldFireWhenAnEdgeIsAdded() {
        var graph = new Graph();

        var handler = (0, _helpers.spy)(function (data) {
          _assert2.default.strictEqual(data.key, 'J->T');
          _assert2.default.deepEqual(data.attributes, { weight: 1 });
          _assert2.default.strictEqual(data.source, 'John');
          _assert2.default.strictEqual(data.target, 'Thomas');
          _assert2.default.strictEqual(data.undirected, false);
        });

        graph.on('edgeAdded', handler);

        graph.addNodesFrom(['John', 'Thomas']);
        graph.addEdgeWithKey('J->T', 'John', 'Thomas', { weight: 1 });

        (0, _assert2.default)(handler.called);
      }
    },

    'nodeDropped': {
      'it should fire when a node is dropped.': function itShouldFireWhenANodeIsDropped() {
        var graph = new Graph();

        var handler = (0, _helpers.spy)(function (data) {
          _assert2.default.strictEqual(data.key, 'John');
          _assert2.default.deepEqual(data.attributes, { age: 34 });
        });

        graph.on('nodeDropped', handler);

        graph.addNode('John', { age: 34 });
        graph.dropNode('John');

        (0, _assert2.default)(handler.called);
      }
    },

    'edgeDropped': {
      'it should fire when an edge is added.': function itShouldFireWhenAnEdgeIsAdded() {
        var graph = new Graph();

        var handler = (0, _helpers.spy)(function (data) {
          _assert2.default.strictEqual(data.key, 'J->T');
          _assert2.default.deepEqual(data.attributes, { weight: 1 });
          _assert2.default.strictEqual(data.source, 'John');
          _assert2.default.strictEqual(data.target, 'Thomas');
          _assert2.default.strictEqual(data.undirected, false);
        });

        graph.on('edgeDropped', handler);

        graph.addNodesFrom(['John', 'Thomas']);
        graph.addEdgeWithKey('J->T', 'John', 'Thomas', { weight: 1 });
        graph.dropEdge('J->T');

        (0, _assert2.default)(handler.called);
      }
    },

    'cleared': {
      'it should fire when the graph is cleared.': function itShouldFireWhenTheGraphIsCleared() {
        var graph = new Graph();

        var handler = (0, _helpers.spy)();

        graph.on('cleared', handler);

        graph.clear();

        (0, _assert2.default)(handler.called);
      }
    },

    'nodeUpdated': {
      'it should fire when a node is updated.': function itShouldFireWhenANodeIsUpdated() {
        var graph = new Graph();

        var handler = (0, _helpers.spy)(function (payload) {
          var key = payload.key;
          var type = payload.type;
          var meta = payload.meta;


          _assert2.default.strictEqual(key, 'John');

          (0, _assert2.default)(VALID_TYPES.has(type));

          if (type === 'set') {
            _assert2.default.strictEqual(meta.name, 'age');
            _assert2.default.strictEqual(meta.value, 34);
          } else if (type === 'replace') {
            _assert2.default.deepEqual(meta.before, { age: 34 });
            _assert2.default.deepEqual(meta.after, { age: 56 });
          } else {
            _assert2.default.deepEqual(meta.data, { eyes: 'blue' });
          }
        });

        graph.on('nodeUpdated', handler);

        graph.addNode('John');
        graph.setNodeAttribute('John', 'age', 34);
        graph.replaceNodeAttributes('John', { age: 56 });
        graph.mergeNodeAttributes('John', { eyes: 'blue' });

        _assert2.default.strictEqual(handler.times, 3);
      }
    },

    'edgeUpdated': {
      'it should fire when a node is updated.': function itShouldFireWhenANodeIsUpdated() {
        var graph = new Graph();

        var handler = (0, _helpers.spy)(function (payload) {
          var key = payload.key;
          var type = payload.type;
          var meta = payload.meta;


          _assert2.default.strictEqual(key, 'J->T');

          (0, _assert2.default)(VALID_TYPES.has(type));

          if (type === 'set') {
            _assert2.default.strictEqual(meta.name, 'weight');
            _assert2.default.strictEqual(meta.value, 34);
          } else if (type === 'replace') {
            _assert2.default.deepEqual(meta.before, { weight: 34 });
            _assert2.default.deepEqual(meta.after, { weight: 56 });
          } else {
            _assert2.default.deepEqual(meta.data, { type: 'KNOWS' });
          }
        });

        graph.on('edgeUpdated', handler);

        graph.addNodesFrom(['John', 'Thomas']);
        graph.addEdgeWithKey('J->T', 'John', 'Thomas');
        graph.setEdgeAttribute('J->T', 'weight', 34);
        graph.replaceEdgeAttributes('J->T', { weight: 56 });
        graph.mergeEdgeAttributes('J->T', { type: 'KNOWS' });

        _assert2.default.strictEqual(handler.times, 3);
      }
    }
  };
}