'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = edgesIteration;

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _helpers = require('../helpers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; } /**
                                                                                                                                                                                                                   * Graphology Edges Iteration Specs
                                                                                                                                                                                                                   * =================================
                                                                                                                                                                                                                   *
                                                                                                                                                                                                                   * Testing the edges iteration-related methods of the graph.
                                                                                                                                                                                                                   */


var METHODS = ['edges', 'inEdges', 'outEdges', 'inboundEdges', 'outboundEdges', 'directedEdges', 'undirectedEdges'];

var SELF_LOOPS_METHODS = ['selfLoops'];

var ALL_METHODS = METHODS.concat(SELF_LOOPS_METHODS);

function edgesIteration(Graph, checkers) {
  var invalid = checkers.invalid;
  var notFound = checkers.notFound;


  var graph = new Graph(null, { multi: true });

  graph.addNodesFrom(['John', 'Thomas', 'Martha', 'Roger', 'Catherine', 'Alone', 'Forever']);

  graph.addDirectedEdgeWithKey('J->T', 'John', 'Thomas');
  graph.addDirectedEdgeWithKey('J->M', 'John', 'Martha');
  graph.addDirectedEdgeWithKey('C->J', 'Catherine', 'John');

  graph.addUndirectedEdgeWithKey('M<->R', 'Martha', 'Roger');
  graph.addUndirectedEdgeWithKey('M<->J', 'Martha', 'John');
  graph.addUndirectedEdgeWithKey('J<->R', 'John', 'Roger');
  graph.addUndirectedEdgeWithKey('T<->M', 'Thomas', 'Martha');

  var ALL_EDGES = ['J->T', 'J->M', 'C->J', 'M<->R', 'M<->J', 'J<->R', 'T<->M'];

  var ALL_DIRECTED_EDGES = ['J->T', 'J->M', 'C->J'];

  var ALL_UNDIRECTED_EDGES = ['M<->R', 'M<->J', 'J<->R', 'T<->M'];

  var TEST_DATA = {
    edges: {
      all: ALL_EDGES,
      node: {
        key: 'John',
        edges: ['C->J', 'J->T', 'J->M', 'M<->J', 'J<->R']
      },
      bunch: {
        keys: ['Martha', 'Roger'],
        edges: ['J->M', 'M<->R', 'M<->J', 'T<->M', 'J<->R']
      },
      path: {
        source: 'John',
        target: 'Martha',
        edges: ['J->M', 'M<->J']
      }
    },
    inEdges: {
      all: ALL_DIRECTED_EDGES,
      node: {
        key: 'John',
        edges: ['C->J']
      },
      bunch: {
        keys: ['Thomas', 'John'],
        edges: ['J->T', 'C->J']
      },
      path: {
        source: 'John',
        target: 'Martha',
        edges: ['J->M']
      }
    },
    outEdges: {
      all: ALL_DIRECTED_EDGES,
      node: {
        key: 'John',
        edges: ['J->T', 'J->M']
      },
      bunch: {
        keys: ['John', 'Catherine'],
        edges: ['J->T', 'J->M', 'C->J']
      },
      path: {
        source: 'John',
        target: 'Martha',
        edges: ['J->M']
      }
    },
    inboundEdges: {
      all: ALL_EDGES,
      node: {
        key: 'John',
        edges: ['C->J', 'M<->J']
      },
      bunch: {
        keys: ['Thomas', 'John'],
        edges: ['J->T', 'C->J', 'M<->J']
      },
      path: {
        source: 'John',
        target: 'Martha',
        edges: ['J->M', 'M<->J']
      }
    },
    outboundEdges: {
      all: ALL_EDGES,
      node: {
        key: 'John',
        edges: ['J->T', 'J->M', 'J<->R']
      },
      bunch: {
        keys: ['John', 'Martha'],
        edges: ['J->T', 'J->M', 'J<->R', 'M<->R', 'M<->J']
      },
      path: {
        source: 'John',
        target: 'Martha',
        edges: ['J->M', 'M<->J']
      }
    },
    directedEdges: {
      all: ALL_DIRECTED_EDGES,
      node: {
        key: 'John',
        edges: ['C->J', 'J->T', 'J->M']
      },
      bunch: {
        keys: ['John', 'Catherine'],
        edges: ['C->J', 'J->T', 'J->M']
      },
      path: {
        source: 'John',
        target: 'Martha',
        edges: ['J->M']
      }
    },
    undirectedEdges: {
      all: ALL_UNDIRECTED_EDGES,
      node: {
        key: 'John',
        edges: ['M<->J', 'J<->R']
      },
      bunch: {
        keys: ['John', 'Martha'],
        edges: ['M<->R', 'M<->J', 'J<->R', 'T<->M']
      },
      path: {
        source: 'John',
        target: 'Martha',
        edges: ['M<->J']
      }
    }
  };

  var graphWithSelfLoops = new Graph(null, { multi: true });
  graphWithSelfLoops.addNodesFrom(['John', 'Tabitha', 'Marcia', 'Alone']);
  graphWithSelfLoops.addEdgeWithKey('J->T', 'John', 'Tabitha');
  graphWithSelfLoops.addEdgeWithKey('J1', 'John', 'John');
  graphWithSelfLoops.addEdgeWithKey('J2', 'John', 'John');
  graphWithSelfLoops.addEdgeWithKey('T1', 'Tabitha', 'Tabitha');
  graphWithSelfLoops.addEdgeWithKey('M1', 'Marcia', 'Marcia');

  var SELF_LOOPS_TEST_DATA = {
    selfLoops: {
      all: ['J1', 'J2', 'T1', 'M1'],
      node: {
        key: 'John',
        loops: ['J1', 'J2']
      },
      bunch: {
        keys: ['John', 'Tabitha'],
        edges: ['J1', 'J2', 'T1']
      }
    }
  };

  function commonTests(name) {
    return _defineProperty({}, '#.' + name, {
      'it should throw if too many arguments are provided.': function itShouldThrowIfTooManyArgumentsAreProvided() {
        _assert2.default.throws(function () {
          graph[name](1, 2, 3);
        }, invalid());
      },

      'it should throw when the node is not found.': function itShouldThrowWhenTheNodeIsNotFound() {
        _assert2.default.throws(function () {
          graph[name]('Test');
        }, notFound());
      },

      'it should throw if any of the provided bunch node is not found.': function itShouldThrowIfAnyOfTheProvidedBunchNodeIsNotFound() {
        _assert2.default.throws(function () {
          graph[name](['Test']);
        }, notFound());
      },

      'it should throw if either source or target is not found.': function itShouldThrowIfEitherSourceOrTargetIsNotFound() {
        _assert2.default.throws(function () {
          graph[name]('Test', 'Alone');
        }, notFound());

        _assert2.default.throws(function () {
          graph[name]('Alone', 'Test');
        }, notFound());
      }
    });
  }

  function specificTests(name, data) {
    var _ref2;

    var counterName = 'count' + (0, _helpers.capitalize)(name);

    return _ref2 = {}, _defineProperty(_ref2, '#.' + name, {
      'it should return all the relevant edges.': function itShouldReturnAllTheRelevantEdges() {
        var edges = graph[name]();

        _assert2.default.deepEqual(edges, data.all);
      },

      'it should return a node\'s relevant edges.': function itShouldReturnANodeSRelevantEdges() {
        var edges = graph[name](data.node.key);

        _assert2.default.deepEqual(edges, data.node.edges);
        _assert2.default.deepEqual(graph[name]('Alone'), []);
      },

      'it should return a bunch of nodes\' relevant edges.': function itShouldReturnABunchOfNodesRelevantEdges() {
        (0, _helpers.testBunches)(data.bunch.keys, function (bunch) {
          var edges = graph[name](bunch);

          (0, _assert2.default)((0, _helpers.sameMembers)(edges, data.bunch.edges));
          _assert2.default.deepEqual(graph[name](['Forever', 'Alone']), []);
        });
      },

      'it should return all the relevant edges between source & target.': function itShouldReturnAllTheRelevantEdgesBetweenSourceTarget() {
        var edges = graph[name](data.path.source, data.path.target);

        (0, _assert2.default)((0, _helpers.sameMembers)(edges, data.path.edges));
        _assert2.default.deepEqual(graph[name]('Forever', 'Alone'), []);
      }
    }), _defineProperty(_ref2, '#.' + counterName, {
      'it should count all the relevant edges.': function itShouldCountAllTheRelevantEdges() {
        var nb = graph[counterName]();

        _assert2.default.strictEqual(nb, data.all.length);
      },

      'it should count all the relevant edges of a node.': function itShouldCountAllTheRelevantEdgesOfANode() {
        var nb = graph[counterName](data.node.key);

        _assert2.default.strictEqual(nb, data.node.edges.length);
        _assert2.default.deepEqual(graph[counterName]('Alone'), 0);
      },

      'it should count a bunch of nodes\' relevant edges.': function itShouldCountABunchOfNodesRelevantEdges() {
        (0, _helpers.testBunches)(data.bunch.keys, function (bunch) {

          _assert2.default.strictEqual(graph[counterName](bunch), data.bunch.edges.length);
          _assert2.default.deepEqual(graph[counterName](['Forever', 'Alone']), 0);
        });
      },

      'it should count all the relevant edges between source & target.': function itShouldCountAllTheRelevantEdgesBetweenSourceTarget() {
        var nb = graph[counterName](data.path.source, data.path.target);

        _assert2.default.strictEqual(nb, data.path.edges.length);
        _assert2.default.deepEqual(graph[counterName]('Forever', 'Alone'), 0);
      }
    }), _ref2;
  }

  function selfLoopsTests(name, data) {
    var _ref3;

    var counterName = 'count' + (0, _helpers.capitalize)(name);

    return _ref3 = {}, _defineProperty(_ref3, '#.' + name, {
      'it should return all the relevant edges.': function itShouldReturnAllTheRelevantEdges() {
        var edges = graphWithSelfLoops[name]();

        _assert2.default.deepEqual(edges, data.all);
      },

      'it should return a node\'s relevant self-loops.': function itShouldReturnANodeSRelevantSelfLoops() {
        var edges = graphWithSelfLoops[name](data.node.key);

        _assert2.default.deepEqual(edges, data.node.loops);
        _assert2.default.deepEqual(graphWithSelfLoops[name]('Alone'), []);
      },

      'it should return a bunch of nodes\' relevant edges.': function itShouldReturnABunchOfNodesRelevantEdges() {
        (0, _helpers.testBunches)(data.bunch.keys, function (bunch) {
          var edges = graphWithSelfLoops[name](bunch);

          (0, _assert2.default)((0, _helpers.sameMembers)(edges, data.bunch.edges));
          _assert2.default.deepEqual(graphWithSelfLoops[name](['Alone']), []);
        });
      }
    }), _defineProperty(_ref3, '#.' + counterName, {
      'it should count all the relevant edges.': function itShouldCountAllTheRelevantEdges() {
        var nb = graphWithSelfLoops[counterName]();

        _assert2.default.strictEqual(nb, data.all.length);
      },

      'it should count all the relevant self-loops of a node.': function itShouldCountAllTheRelevantSelfLoopsOfANode() {
        var nb = graphWithSelfLoops[counterName](data.node.key);

        _assert2.default.strictEqual(nb, data.node.loops.length);
        _assert2.default.deepEqual(graphWithSelfLoops[counterName]('Alone'), 0);
      },

      'it should count a bunch of nodes\' relevant edges.': function itShouldCountABunchOfNodesRelevantEdges() {
        (0, _helpers.testBunches)(data.bunch.keys, function (bunch) {

          _assert2.default.strictEqual(graphWithSelfLoops[counterName](bunch), data.bunch.edges.length);
          _assert2.default.deepEqual(graphWithSelfLoops[counterName](['Alone']), 0);
        });
      }
    }), _ref3;
  }

  var tests = {};

  // Common tests
  ALL_METHODS.forEach(function (name) {
    return (0, _helpers.deepMerge)(tests, commonTests(name));
  });
  ALL_METHODS.forEach(function (name) {
    return (0, _helpers.deepMerge)(tests, commonTests('count' + (0, _helpers.capitalize)(name)));
  });

  // Specific tests
  for (var name in TEST_DATA) {
    (0, _helpers.deepMerge)(tests, specificTests(name, TEST_DATA[name]));
  }for (var _name in SELF_LOOPS_TEST_DATA) {
    (0, _helpers.deepMerge)(tests, selfLoopsTests(_name, SELF_LOOPS_TEST_DATA[_name]));
  }return tests;
}